#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#include <libgen.h>
#include <pthread.h>
//#include <sys/time.h>  
#include <sys/ipc.h>
#include <sys/sem.h>

#include <adwin-macros.h>
#include <linux/adwin.h>
#include "libadwin.h"
#include "libadwin/errno.h"
#include "libadwin/types.h"
#include "configfile.h"

#ifdef WITH_ETHERNET
	#include <sys/socket.h>
	#include <sys/un.h>
	#include <netinet/in.h>
	#include <arpa/inet.h>
	#include <sys/time.h>
	#define USE_FOR_LIBADWIN 1
	#include "ADEthernet.h"	/* Generischer Treiber fr Ethernet */
//  #include "ethernet.h" 	/* Prototypen fr Netz-Funktionen (NET_*) */

	#define Generic_Driver_Get_Par_All Generic_Driver_Get_Par_Block
	#define Generic_Driver_Get_FPar_All Generic_Driver_Get_FPar_Block
#endif
#ifdef WITH_LINK
  #include "transputer.h" 	/* Prototypen fr Transputer-Funktionen (T_*) */
  #include "dsp.h" 		/* Prototypen fr DSP-Funktionen (DSP_*) */
#endif
#ifdef WITH_USB
  #include "usb.h" 		/* Prototypen fr USB-Funktionen (USB_*) */
#endif

#include "libadwin-macros.h"

#define g_QNAN_MASKE 0x7F800000

const long g_Version = 2000015;

// used to initialize the shared memory, and load config file
int initialized = 0;

/* Thread-specific errno (begin) */

static pthread_key_t adwin_errno_key;
static pthread_once_t adwin_errno_once = PTHREAD_ONCE_INIT;

static void __adwin_errno_destructor(void *ptr) {
	free(ptr);
}

static void __adwin_errno_once(void) {
	pthread_key_create(&adwin_errno_key, __adwin_errno_destructor);
}

inline int *__adwin_errno_location(void) {
	int *p;
	pthread_once(&adwin_errno_once, __adwin_errno_once);

	if ((p = pthread_getspecific(adwin_errno_key)) == NULL) {
		p = malloc(sizeof(int));
		*p = 0;
		pthread_setspecific(adwin_errno_key, p);
	}

	return p;
}

/* Thread-specific errno (end) */

#include "labview.c"	// This file contains wrapper functions which are 
                        // used only for LABview to get the correct names

typedef struct _DeviceNumber
{
	int DeviceNo;
	int DeviceType;
} DeviceNumber;

/* DeviceNo fr einzelne Threads */
static pthread_key_t adwin_deviceno_key;
static pthread_once_t adwin_deviceno_once = PTHREAD_ONCE_INIT;

static void __adwin_deviceno_destructor(void *ptr) {
	free(ptr);
}

static void __adwin_deviceno_once(void) {
	pthread_key_create(&adwin_deviceno_key, __adwin_deviceno_destructor);
}

static inline DeviceNumber *__adwin_deviceno_location(void) {
	DeviceNumber *p;
	pthread_once(&adwin_deviceno_once, __adwin_deviceno_once);

	if ((p = pthread_getspecific(adwin_deviceno_key)) == NULL) {
		p = (DeviceNumber*) malloc (sizeof (DeviceNumber));
		p->DeviceType = 0;			// Link
		p->DeviceNo = 0;			// Nicht gesetzt
		pthread_setspecific(adwin_deviceno_key, p);
	}

	return p;
}

#define CURRENTDEVICE (*(__adwin_deviceno_location ()))

/* Ende Thread-spezifische DeviceNo */

/***********************************************************************
 * Semaphores 
 */
// global variables for semaphore locking

int semID_glob, semID_dev;

static inline void init_glob_semaphore() {
	semID_glob = semget(SEM_KEY + 1, 1, 0666 | IPC_CREAT | IPC_EXCL);
	
	if (semID_glob >= 0) { 
		// it worked -- we are the first library instance
		DEBUG_SEM("global semaphore created at %d\n", semID_glob);
		// initialize semaphore 
		union semun sem_init;
		unsigned short values[1];
		values[0] = 1; 
		sem_init.array = values; 
		semctl(semID_glob, 0, SETALL, sem_init);
	} else if (errno==EEXIST) {
		// the semaphore set already exists, so just use it
		if ((semID_glob = semget(SEM_KEY + 1, 1, 0666)) < 0) {
			perror("semget global");
		}
		DEBUG_SEM("using existing global semaphore %d\n", semID_glob);
	} else {
		perror("semget global");
	}
}


static inline void grab_glob_semaphore() {
	struct sembuf sem = {0, -1, SEM_UNDO};
	
	if (semop(semID_glob, &sem, 1) == -1) {
		perror("grab global semaphore");
	}
	DEBUG_SEM("Grabbed global semaphore, value is %d\n", 
		   semctl(semID_glob, 0, GETVAL, NULL));
}

static inline void release_glob_semaphore() {
	struct sembuf sem = {0, 1, SEM_UNDO};

	if (semop(semID_glob, &sem, 1) == -1) {
		perror("release global semaphore");
	}
	DEBUG_SEM("Released global semaphore, value is %d\n", 
		   semctl(semID_glob, 0, GETVAL, NULL));
}

static inline void init_dev_semaphore() {
	// try to create new semaphore set
	semID_dev = semget(SEM_KEY, MAX_DEVICES, 0666 | IPC_CREAT | IPC_EXCL);
	if (semID_dev >= 0) { 
		// it worked -- we are the first library instance
		DEBUG_SEM("device semaphores created at %d\n", semID_dev);
		// initialize semaphore set
		union semun sem_init;
		unsigned short values[MAX_DEVICES];
		int i;
		for (i = 0; i < MAX_DEVICES; i++)
			values[i] = 1; 
		sem_init.array = values; 
		semctl(semID_dev, 0, SETALL, sem_init);
	} else if (errno==EEXIST) {
		// the semaphore set already exists, so just use it
		if ((semID_dev = semget(SEM_KEY, 1, 0666)) < 0) {
				perror("semget devices");
		}
		DEBUG_SEM("using existing device semaphores %d\n", semID_dev);
	} else {
		perror("semget global");
	}
}

static inline void grab_dev_semaphore(int index) {
	struct sembuf sem = {index, -1, SEM_UNDO};
	
	if (semop(semID_dev, &sem, 1) == -1) {
		perror("grab device semaphore");
	}
	DEBUG_SEM("Grabbed device semaphore at index %d, value is %d\n", 
		   index, 
		   semctl(semID_dev, index, GETVAL, NULL));
}

static inline void release_dev_semaphore(int index) {
	struct sembuf sem = {index, 1, SEM_UNDO };

	if (semop(semID_dev, &sem, 1) == -1) {
		perror("release semaphore");
	}
	DEBUG_SEM("Released device semaphore at index %d, value is %d\n", 
		   index,
		   semctl(semID_dev, index, GETVAL, NULL));
}


/*****************************************************************************/
static inline int SETMUTEX(int device, adwin_dev_t *vd) {
	int dev;

#ifdef KERNEL_MODULE
	char file[80];

	sprintf(file, "/proc/adwin/dev/0x%X", device);
	if ((dev = open(file, O_RDWR)) < 0) {
		if ((errno == ENOENT) || (errno == ENODEV))
			adwin_errno = k_this_link_is_unknown;
		
		return -1;
	}

	ioctl(dev, ADIOGKDEVINFO, vd);
#else
#define BIGNUM 16384
	if (! initialized) { // library has just been loaded - setup semaphores
		             // and read config file
		init_dev_semaphore();
		init_glob_semaphore();
		grab_glob_semaphore();
		reloadConfigFile();
		release_glob_semaphore();
		initialized = 1;
	}
	
	if ((dev = findDevice(device)) == -1) { // not found
		PRINTDEBUG("device %d not found\n", device);
		adwin_errno = k_this_link_is_unknown;
		clearDevice(vd);
		return -1;
	}
	grab_dev_semaphore(dev);
	*vd = device_list[dev];
	//	printDeviceConfig(stderr, *vd);
#endif

	switch(vd->dev_type) {
#if defined(WITH_LINK)
		case DC_LINK:
		case DC_PCMCIA:
			return dev;
#endif
#if defined(WITH_USB)
		case DC_USB:
			adwin_errno = -ENOTSUP;
			return -1;
#endif
#if defined(WITH_ETHERNET)
		case DC_ETHERNET:
			if (vd->dev_processor != DP_DSP)
				vd->dev_processor = DP_DSP;

/*			if (! vd->dev_server_version) {
				vd->dev_processor = DP_DSP;
				if((vd->dev_server_version = __get_Version_TcpIp(vd)) < 0) {
					adwin_errno = k_wrong_version;
					return -1;
				}
			}*/
			return dev;
#endif
		default:	// DC_UNKNOWN:
			adwin_errno = k_unknown_device;
			return -1;
	}
}

static inline int SETMUTEX_INIT_GEN_DRIVER (int device, adwin_dev_t *vd)
{
	Set_DeviceNo(device);
	return SETMUTEX (device, vd);
}

/*****************************************************************************/
static inline void CLEARMUTEX(adwin_dev_t *vd, int fd) {

#ifndef KERNEL_MODULE
	release_dev_semaphore(fd);
	return;
#endif

	switch(vd->dev_type) {
#if defined (WITH_LINK)
		case DC_LINK:
		case DC_PCMCIA:
   			close(fd);
   			break;
#endif
#if defined(WITH_ETHERNET)
		case DC_ETHERNET:
			close(fd);
			break;
#endif
#if defined(WITH_USB)
		case DC_USB:
			break;
#endif
		default:
			return;
	}
}

#include "libadwin-macros.h"

void Set_DeviceNo (int Number)
{
	adwin_dev_t Dev;
	int fd;
	DeviceNumber dn;
	DeviceNumber NewDn;

	if ((Number < 1) || (Number > 0xfff)) {
		adwin_errno = k_this_link_not_allowed;
		return;
	}

	memcpy (&dn, &CURRENTDEVICE, sizeof (DeviceNumber));
	if (dn.DeviceNo == Number) return;

	if ((fd = SETMUTEX (Number, &Dev)) >= 0)
	{
		if (Dev.dev_type == DC_ETHERNET)
			if (dn.DeviceType == DC_ETHERNET)
				reConnect (inet_ntoa (Dev.dev_net_server.sin_addr),
					   Dev.dev_net_host_password,
					   Dev.dev_net_host_link,
					   ntohs (Dev.dev_net_server.sin_port),
					   Dev.dev_net_timeout);
			else
				openConnect (inet_ntoa (Dev.dev_net_server.sin_addr),
					   Dev.dev_net_host_password,
					   Dev.dev_net_host_link,
					   ntohs (Dev.dev_net_server.sin_port),
					   Dev.dev_net_timeout);
		else
			if (dn.DeviceType == DC_ETHERNET)
				closeConnect ();

		NewDn.DeviceNo = Number;
		NewDn.DeviceType = Dev.dev_type;
		memcpy (&CURRENTDEVICE, &NewDn, sizeof (DeviceNumber));

		CLEARMUTEX (&Dev, fd);
		adwin_errno = 0;
		return;
	}
	
	if (adwin_errno == k_this_link_is_unknown)
		return;
	adwin_errno = k_unknown_device;
}


/*****************************************************************************/
void Boot(const char *Filename)
{
	long res;

	res = adwin_boot(Filename, 100000, CURRENTDEVICE.DeviceNo);

	if (res > 1000)
	{
		if (res != 100000)
		{
			if (res != 8000)
				adwin_boot(Filename, res, CURRENTDEVICE.DeviceNo);
		}
	}
}

long adwin_boot(const char *btl, long iboardsize, int id) {
	adwin_dev_t dev;
	int fd;
	long tmp = 1;
	struct stat s;
	char share[PATH_MAX];

	if (stat(btl, &s) < 0) {
		if (! getAdwindir(share)) {
			adwin_errno = -errno;
			if (adwin_errno == -ENOENT) adwin_errno = k_File_not_found;
			return 1;
		}

		strcat(share, "/share/btl/");
		strcat(share, basename((char *)btl));

		if (stat(share, &s) < 0) {
			adwin_errno = -errno;
			if (adwin_errno == -ENOENT) adwin_errno = k_File_not_found;
			return 1;
		} else
			btl = share;
	}

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return 1;

	switch(dev.dev_type) {
#ifdef WITH_ETHERNET
		case DC_ETHERNET:
			Generic_Driver_Boot((char *)btl);
			tmp = (adwin_errno = Generic_Driver_Get_Last_Error ()) ? -1 : 0;
			break;
#endif
#ifdef WITH_USB
		case DC_USB:
			tmp = USB_Boot(&dev, fd, (char *)btl, iboardsize);
			break;
#endif
#if defined(WITH_LINK)
		case DC_LINK:
		case DC_PCMCIA:
			/*
			 * Am BTL-Namen determinieren welcher Prozessor
			 */
//			SIGSEV ??
//			for (i = 0; i < (signed)strlen(btl); i++)
//				btl[i] = tolower(btl[i]);

			if (strstr(btl, "adwin9.btl") != NULL) {
				tmp = DSP_Boot(&dev, fd, (char *)btl, iboardsize);
			} else {
				tmp = T_Boot(&dev, fd, (char *)btl, iboardsize);
			}
			break;
#endif
		default:
			return 1;
	}

	CLEARMUTEX(&dev, fd);

	return tmp == -1 ? 1 : tmp;
}



/*****************************************************************************/
/*
long AD_Memory(int id = Nummber;id) {
	adwin_dev_t dev;
	int fd, tmp = 0;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, AD_Memory, &dev, fd);

	CLEARMUTEX(&dev, fd);
	return (tmp > -1) ? tmp : 0;
}
*/

/*****************************************************************************/
long Free_Mem (long Mem_Spec){
	return adwin_mem(Mem_Spec, CURRENTDEVICE.DeviceNo);
}

long adwin_mem(int typ, int id) {
	int err, memsize;
	memsize = __adwin_mem(typ, &err, id);
	return (err == -1 ? 255 : memsize);
}

long __adwin_mem(int typ, int *err, int id) {
	adwin_dev_t dev;
	int fd, tmp = -1;
	int32_t res;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL_WITH_RETURN1 (tmp, Free_Mem, &dev, fd, typ, &res);

	// don't multiply for an ethernet device, as the generic driver already returns the right values
	if (((dev.dev_processor == DP_DSP) || (dev.dev_processor == DP_21160)) && dev.dev_type != DC_ETHERNET) {
		switch (typ) {
			case 1:	res *= 6; 	break;		// PM
			case 3:	res *= 4; 	break;		// DM
			case 4:	res *= 4; 	break;		// DRAM

			default:
				adwin_errno = -EINVAL;
				tmp = -1;
				res = 0;
		}
	}
	CLEARMUTEX(&dev, fd);

	if (err)
		*err = adwin_errno;

	return res;
}


/*****************************************************************************/
/*int adwin_get_memory_dsp(long base_addr, long anzahl, int32_t *dest, int id) {
	adwin_dev_t dev;
	int fd, tmp = -1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Get_Memory_dsp, &dev, fd, base_addr, anzahl, dest);

	CLEARMUTEX(&dev, fd);
	return tmp;
}
*/
/*****************************************************************************/
long Test_Version (){
	return adwin_test_version (CURRENTDEVICE.DeviceNo);
}

int adwin_test_version (int id) {
	adwin_dev_t dev;
	int fd, tmp = 3;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL_WITH_RETURN0(tmp, Test_Version, &dev, fd);
	CLEARMUTEX(&dev, fd);

	return tmp;
}


/*****************************************************************************/
void Load_Process(const char *Filename) {
	adwin_process_load(Filename, CURRENTDEVICE.DeviceNo);
}

int adwin_process_load(const char *dateiname, int id) {
	adwin_dev_t dev;
	int fd, tmp = -1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Load_Process, &dev, fd, (char *)dateiname);

	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 1;
}

long Process_Status (long ProcessNo) {
	return adwin_process_status (ProcessNo, CURRENTDEVICE.DeviceNo);
}

int adwin_process_status (int ProcessNo, int id) {
	return adwin_get_par(ProcessNo - 100, id);
}

/*****************************************************************************/
long Get_Par(long Index) {
	return adwin_get_par(Index, CURRENTDEVICE.DeviceNo);
}


int adwin_get_par(int parID, int id) {
	int err, value;
	value = __adwin_get_par(parID, &err, id);

	return ( err == -1) ? 255 : value;
}

/* Linux-only */
int __adwin_get_par(int parID, int *err, int id) {
	adwin_dev_t dev;
	int fd, tmp = -1;
	int32_t dest = -1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

//	SELECT_PROTOCOL_WITH_RETURN1(tmp, Get_Par, &dev, fd, parID + 1000, &dest);
	switch (dev.dev_type)
	{
#ifdef WITH_LINK
		case DC_LINK:
			switch (dev.dev_processor)
			{
				case DP_DSP:
					tmp = DSP_Get_Par (&dev, fd, 1000 + parID, &dest);
				break;
				case DP_T2:
				case DP_T4:
				case DP_T8:
					tmp = T_Get_Par (&dev, fd, 1000 + parID, &dest);
			}
		break;
#endif
#ifdef WITH_USB
		case DC_UDB:
			tmp = USB_Get_Par (&dev, fd, 1000 + parID, &dest);
		break;
#endif
#ifdef WITH_ETHERNET
		case DC_ETHERNET:
		{
			dest = Generic_Driver_Get_Par (parID);
			tmp = adwin_errno = Generic_Driver_Get_Last_Error ();
		}
#endif
	}
	CLEARMUTEX(&dev, fd);

	if (err)
		*err = tmp;

	return dest;
}

/*****************************************************************************/
float Get_FPar(long Index) {
  return adwin_get_fpar(Index, CURRENTDEVICE.DeviceNo);
}

float adwin_get_fpar(int parID, int id) {
	int err;
	float value = __adwin_get_fpar(parID, &err, id);

	return (err == -1) ? 255.0 : value;
}

float  __adwin_get_fpar(int parID, int *err, int id) {
	adwin_dev_t dev;
	int fd, tmp = -1;
	float dest = -1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

//	SELECT_PROTOCOL_WITH_RETURN1(tmp, Get_FPar, &dev, fd, parID + 1100, &dest);
	switch (dev.dev_type)
	{
#ifdef WITH_LINK
		case DC_LINK:
			switch (dev.dev_processor)
			{
				case DP_DSP:
					tmp = DSP_Get_FPar (&dev, fd, 1100 + parID, &dest);
				break;
				case DP_T2:
				case DP_T4:
				case DP_T8:
					tmp = T_Get_FPar (&dev, fd, 1100 + parID, &dest);
			}
		break;
#endif
#ifdef WITH_USB
		case DC_UDB:
			tmp = USB_Get_FPar (&dev, fd, 1100 + parID, &dest);
		break;
#endif
#ifdef WITH_ETHERNET
		case DC_ETHERNET:
		{
			dest = Generic_Driver_Get_FPar (parID);
			tmp = adwin_errno = Generic_Driver_Get_Last_Error ();
		}
#endif
	}
	CLEARMUTEX(&dev, fd);

	if (err)
		*err = tmp;

	return dest;
}

/*****************************************************************************/
void Set_Par(long Index, long Value){
  adwin_set_par(Index, Value, CURRENTDEVICE.DeviceNo);
}

int adwin_set_par(int parID, int value, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

//	SELECT_PROTOCOL(tmp, Set_Par, &dev, fd, parID + 1000, value);
	switch (dev.dev_type)
	{
#ifdef WITH_LINK
		case DC_LINK:
			switch (dev.dev_processor)
			{
				case DP_DSP:
					tmp = DSP_Set_Par (&dev, fd, 1000 + parID, value);
				break;
				case DP_T2:
				case DP_T4:
				case DP_T8:
					tmp = T_Set_Par (&dev, fd, 1000 + parID, value);
			}
		break;
#endif
#ifdef WITH_USB
		case DC_UDB:
			tmp = USB_Set_Par (&dev, fd, 1000 + parID, value);
		break;
#endif
#ifdef WITH_ETHERNET
		case DC_ETHERNET:
		{
			Generic_Driver_Set_Par (parID, value);
			tmp = adwin_errno = Generic_Driver_Get_Last_Error ();
		}
#endif
	}

	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/
void Set_FPar(long Index, float Value){
  adwin_set_fpar(Index, Value, CURRENTDEVICE.DeviceNo);
}

int adwin_set_fpar(int parID, float value, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

//	SELECT_PROTOCOL(tmp, Set_FPar, &dev, fd, parID + 1100, value);
	switch (dev.dev_type)
	{
#ifdef WITH_LINK
		case DC_LINK:
			switch (dev.dev_processor)
			{
				case DP_DSP:
					tmp = DSP_Set_FPar (&dev, fd, 1100 + parID, value);
				break;
				case DP_T2:
				case DP_T4:
				case DP_T8:
					tmp = T_Set_FPar (&dev, fd, 1100 + parID, value);
			}
		break;
#endif
#ifdef WITH_USB
		case DC_UDB:
			tmp = USB_Set_FPar (&dev, fd, 1100 + parID, value);
		break;
#endif
#ifdef WITH_ETHERNET
		case DC_ETHERNET:
		{
			Generic_Driver_Set_FPar (parID, value);
			tmp = adwin_errno = Generic_Driver_Get_Last_Error ();
		}
#endif
	}

	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 0;
}


/*****************************************************************************/
void Get_Par_All(long *Array){
	adwin_get_par_all ((int32_t*)Array, CURRENTDEVICE.DeviceNo);
}


int adwin_get_par_all (int32_t *dest, int id) {
	return adwin_get_par_block (1, 80, dest, id);
}

/*****************************************************************************/
void Get_Par_Block (long *Array, long StartIndex, long Count){
	adwin_get_par_block (StartIndex, Count, (int32_t *)Array, CURRENTDEVICE.DeviceNo);
}



int adwin_get_par_block (int first, int count, int32_t *dest, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL_WITH_RETURN2(tmp, Get_Par_All, &dev, fd, first, count, (long*)dest);

	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/
void Get_FPar_All(float *Array){
	adwin_get_fpar_all (Array, CURRENTDEVICE.DeviceNo);
}


int adwin_get_fpar_all (float *dest, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	if (dev.dev_type == DC_ETHERNET)
	{
		SELECT_PROTOCOL_WITH_RETURN2(tmp, Get_FPar_All, &dev, fd, 1, 80, dest);
	}
	else
	{
		SELECT_PROTOCOL_WITH_RETURN2(tmp, Get_FPar_All, &dev, fd, 101, 80, dest);
	}

	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/
void Get_FPar_Block (float *Array, long StartIndex, long Count){
	adwin_get_fpar_block (StartIndex, Count, Array, CURRENTDEVICE.DeviceNo);
}

int adwin_get_fpar_block (int first, int count, float *dest, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	if (dev.dev_type != DC_ETHERNET)
	{
		SELECT_PROTOCOL_WITH_RETURN2(tmp, Get_FPar_All, &dev, fd, first + 100, count, dest);
	}
	else
	{
		SELECT_PROTOCOL_WITH_RETURN2(tmp, Get_FPar_All, &dev, fd, first, count, dest);
	}
	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/
void Get_FPar_All_Double (double *Array){
	adwin_get_fpar_all_double (Array, CURRENTDEVICE.DeviceNo);
}

int adwin_get_fpar_all_double (double *dest, int id) {
	int i, tmp = 255;

	float fbuf[80] = { 0 };

	tmp = adwin_get_fpar_block (1, 80, fbuf, id);

	if (tmp != -1)
		for (i = 0; i < 80; i++)
			dest[i] = (double) fbuf[i];

	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/
void Get_FPar_Block_Double(double *Array, long StartIndex, long Count){
	adwin_get_fpar_block_double (StartIndex, Count, Array, CURRENTDEVICE.DeviceNo);
}

int adwin_get_fpar_block_double (int first, int count, double *dest, int id) {
	int i, tmp = 255;

	float fbuf[80] = { 0 };

	tmp = adwin_get_fpar_block (first, count, fbuf, id);

	if (tmp != -1)
		for (i = 0; i < count - first + 1; i++)
			dest[i] = (double) fbuf[i];

	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/
long Workload(long Priority){
	return adwin_workload(CURRENTDEVICE.DeviceNo);
}

int adwin_workload(int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL_WITH_RETURN0(tmp, Workload, &dev, fd);

	CLEARMUTEX(&dev, fd);
	return (tmp == -1) ? 255 : tmp;
}

/*****************************************************************************/
void Stop_Process(long ProcessNo){
	adwin_process_stop(ProcessNo, CURRENTDEVICE.DeviceNo);
}


int adwin_process_stop(int processID, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

//	SELECT_PROTOCOL(tmp, Stop_Process, &dev, fd, 99 + 2 * processID);
	switch (dev.dev_type)
	{
#ifdef WITH_LINK
		case DC_LINK:
			switch (dev.dev_processor)
			{
				case DP_DSP:
					DSP_Stop_Process (&dev, fd, 99 + 2 * processID);
				break;
				case DP_T2:
				case DP_T4:
				case DP_T8:
					T_Stop_Process (&dev, fd, 99 + 2 * processID);
			}
		break;
#endif
#ifdef WITH_USB
		case DC_UDB:
			USB_Stop_Process (&dev, fd, 99 + 2 * processID);
		break;
#endif
#ifdef WITH_ETHERNET
		case DC_ETHERNET:
			Generic_Driver_Stop_Process (processID);
			adwin_errno = Generic_Driver_Get_Last_Error ();
#endif
	}

	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/
void Start_Process (long ProcessNo){
	adwin_process_start(ProcessNo, CURRENTDEVICE.DeviceNo);
}


int adwin_process_start(int processID, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

//	SELECT_PROTOCOL(tmp, Start_Process, &dev, fd, 99 + 2 * processID);

	switch (dev.dev_type)
	{
#ifdef WITH_LINK
		case DC_LINK:
			switch (dev.dev_processor)
			{
				case DP_DSP:
					DSP_Start_Process (&dev, fd, 99 + 2 * processID);
				break;
				case DP_T2:
				case DP_T4:
				case DP_T8:
					T_Start_Process (&dev, fd, 99 + 2 * processID);
			}
		break;
#endif
#ifdef WITH_USB
		case DC_UDB:
			USB_Start_Process (&dev, fd, 99 + 2 * processID);
		break;
#endif
#ifdef WITH_ETHERNET
		case DC_ETHERNET:
			Generic_Driver_Start_Process (processID);
			adwin_errno = Generic_Driver_Get_Last_Error ();
#endif
	}

	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/
void Clear_Process (long ProcessNo){
	adwin_process_clear(ProcessNo, CURRENTDEVICE.DeviceNo);
}

int adwin_process_clear(int processID, int id) {
	adwin_dev_t dev;
	int fd, tmp = 1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Clear_Process, &dev, fd, processID);

	CLEARMUTEX(&dev, fd);
	return (tmp != -1) ? 0 : 1;
}


/*****************************************************************************/
long Data_Length (long DataNo){
	return adwin_data_length (DataNo, CURRENTDEVICE.DeviceNo);
}

int adwin_data_length (int dataID, int id) {
	int err, value;
	value = __adwin_data_length(dataID, &err, id);

	return ( err == -1) ? 255 : value;
}

/* Linux-only */
int __adwin_data_length(int dataID, int *err, int id) {
	adwin_dev_t dev;
	int fd, tmp = -1;
	int32_t dest = -1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	switch (dev.dev_type)
	{
#ifdef WITH_LINK
		case DC_LINK:
			switch (dev.dev_processor)
			{
				case DP_DSP:
					tmp = DSP_Data_Length (&dev, fd, dataID);
				break;
				case DP_T2:
				case DP_T4:
				case DP_T8:
					tmp = T_Get_Par (&dev, fd, dataID, &dest);
			}
		break;
#endif
#ifdef WITH_USB
		case DC_UDB:
			tmp = USB_Get_Par (&dev, fd, dataID, &dest);
		break;
#endif
#ifdef WITH_ETHERNET
		case DC_ETHERNET:
		{
			dest = Generic_Driver_Data_Length (dataID);
			tmp = adwin_errno = Generic_Driver_Get_Last_Error ();
		}
#endif
	}
	CLEARMUTEX(&dev, fd);

	if (err)
		*err = tmp;

	return dest;
}

/*****************************************************************************/

int adwin_clear_data (int dataID, int id) {
	adwin_dev_t dev;
	int fd, tmp = 1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Clear_Data, &dev, fd, dataID);

	CLEARMUTEX(&dev, fd);
	return tmp;
}

/*****************************************************************************/
int  adwin_get_data_type(int datanr, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

//	SELECT_PROTOCOL_WITH_RETURN0 (tmp, GetDataType, &dev, fd, datanr);
	switch (dev.dev_processor | dev.dev_type)
	{
		CASE_NET_RET (tmp, GetDataType, datanr)
		CASE_DSP (tmp, GetDataType, &dev, fd, datanr)
		CASE_USB (tmp, GetDataType, &dev, fd, datanr)
		CASE_TRANSPUTER (tmp, GetDataType, &dev, fd, datanr)
		default:
			adwin_errno = -EINVAL;
			tmp = -1;
	}

	CLEARMUTEX(&dev, fd);
	return tmp;
}

/*****************************************************************************/
/*int  CreateData(int datanr, int type, long groesse, int ort, int art, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, CreateData, &dev, fd, datanr, type, groesse, ort, art);

	CLEARMUTEX(&dev, fd);
	return tmp;
}
*/


#define CAST(from_type, from_ptr, to_type, to_ptr, cnt) 		\
do {									\
	int i;								\
	from_type *__fptr = (from_type *) (from_ptr);			\
	to_type   *__tptr = (to_type *) (to_ptr);			\
	for (i = 0; i < (cnt); i++)					\
		*(__tptr)++ = (to_type) *__fptr++;			\
} while(0)

#define FCAST(from_type, from_ptr, to_type, to_ptr, cnt) 		\
do {									\
	int i;								\
	from_type *__fptr = (from_type *) (from_ptr);			\
	to_type   *__tptr = (to_type *) (to_ptr);			\
	for (i = 0; i < (cnt); i++, __tptr++, __fptr++)			\
		if ((*((unsigned int *)__fptr) & g_QNAN_MASKE) != g_QNAN_MASKE)	\
			*(__tptr) = (to_type) *(__fptr);		\
		else							\
			*(__tptr) = (to_type)(0x7FFFFFFFFFFFFFFFLL >> (64 - sizeof (to_type) * 8));	\
} while (0)


int adwin_get_data_int16 (int dataID, long first, long count, int16_t *dest, int id) {
	int res;
	int32_t *ptr;

	ptr = malloc(count * 4);
	res = adwin_get_data_int32(dataID, first, count, ptr, id);
	CAST(int32_t, ptr, int16_t, dest, count);
	free(ptr);
	return res;
}

/*****************************************************************************/
void GetData_Long(long DataNo, long *Data, long StartIndex, long Count){
	adwin_get_data_int32 (DataNo, StartIndex, Count, (int32_t*)Data, CURRENTDEVICE.DeviceNo);
}

int adwin_get_data_int32 (int dataID, long first, long count, int32_t *dest, int id) {
	adwin_dev_t dev;
	int fd, res, data_type;
	void *ptr;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0) {
		return 255;
	}

	switch (dev.dev_processor | dev.dev_type)
	{
		case DP_DSP | DC_ETHERNET:
		case DP_UNKNOWN | DC_ETHERNET:
			data_type = -2;
		break;
		CASE_DSP (data_type, GetDataType, &dev, fd, dataID)
		CASE_USB (data_type, GetDataType, &dev, fd, dataID)
		CASE_TRANSPUTER (data_type, GetDataType, &dev, fd, dataID)
		default:
			adwin_errno = -EINVAL;
			data_type = -1;
	}

	switch(data_type) {
		case 2:
			ptr = malloc(count * 2);
			break;
		case 3:
		case 4:
		case -2:
			ptr = dest;
			break;
		case 5:
			ptr = malloc(count * 4);
			break;

		default:
			adwin_errno = k_Data_is_to_small;
			CLEARMUTEX(&dev, fd);
			return 255;
	}

	if (!ptr ) {
		adwin_errno = -errno;
		return 255;
	}

	if ((data_type == 3 || data_type == 4)
	    && dev.dev_type == DC_ETHERNET)
		data_type = 2;
	SELECT_PROTOCOL_DATA (Get, res, &dev, fd, dataID, first, count, data_type == -2 ? 2 : data_type, ptr);
//	SELECT_PROTOCOL(res, GetData, &dev, fd, dataID, first, count, ptr);
	CLEARMUTEX(&dev, fd);

	switch (data_type) {
		case 2:		/* short */
			if (! res)
				CAST(int16_t, ptr, int32_t, dest, count);
			free(ptr);
			break;
		case 3:		/* int. long, nothing to do */
		case 4:
			break;
		case 5:		/* float to int32 */
			if(! res)
				FCAST(float, ptr, int32_t, dest, count);
			free(ptr);
			break;
	}

	return res ? 255 : 0;
}

/*****************************************************************************/
int adwin_get_data_int64 (int dataID, long first, long count, int64_t *dest, int id) {
	int res;
	int32_t *ptr;

	ptr = malloc(count * 4);
	res = adwin_get_data_int32(dataID, first, count, ptr, id);
	CAST(int32_t, ptr, int64_t, dest, count);
	free(ptr);
	return res;
}

/*****************************************************************************/
int adwin_set_data_int16 (int dataID, long first, long count, int16_t *src, int id) {
	int res;
	int32_t *ptr;

	ptr = malloc(count * 4);
	CAST(int16_t, src, int32_t, ptr, count);
	res = adwin_set_data_int32(dataID, first, count, ptr, id);
	free(ptr);
	return res;
}

/*****************************************************************************/
void SetData_Long(long DataNo, long *Data, long StartIndex, long Count){
	adwin_set_data_int32 (DataNo, StartIndex, Count, (int32_t*)Data, CURRENTDEVICE.DeviceNo);
}

int adwin_set_data_int32 (int dataID, long first, long count, int32_t *src, int id) {
	adwin_dev_t dev;
	int fd, data_type, res;
	void *ptr;

	data_type = adwin_get_data_type(dataID, id);
	switch(data_type) {
		case 2:
			ptr = malloc(count * 2);
			CAST(int32_t, src, int16_t, ptr, count);
			break;
		case 3:
		case 4:
			ptr = src;
			break;
		case 5:
			ptr = malloc(count * 4);
			CAST(int32_t, src, float, ptr, count);
			break;

		default:
			return 255;
	}

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0) {
		if (ptr != src) free(ptr);
		return 255;
	}

	if ((data_type == 3 || data_type == 4)
	    && dev.dev_type == DC_ETHERNET)
		data_type = 2;
	SELECT_PROTOCOL_DATA (Set, res, &dev, fd, dataID, first, count, data_type, ptr);
	CLEARMUTEX(&dev, fd);

	if (ptr != src) free(ptr);
	return res ? 255 : 0;
}

/*****************************************************************************/
int adwin_set_data_int64 (int dataID, long first, long count, int64_t *src, int id) {
	int res;
	int32_t *ptr;

	ptr = malloc(count * 4);
	CAST(int64_t, src, int32_t, ptr, count);
	res = adwin_set_data_int32(dataID, first, count, ptr, id);
	free(ptr);
	return res;
}

/*****************************************************************************/
void GetData_Float(long DataNo, float *Data, long StartIndex, long Count){
	adwin_get_data_float (DataNo, StartIndex, Count, Data, CURRENTDEVICE.DeviceNo);
}

int adwin_get_data_float (int dataID, long first, long count, float *dest, int id) {
	adwin_dev_t dev;
	int fd, res, data_type;
	void *ptr;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0) {
		return 255;
	}

	switch (dev.dev_processor | dev.dev_type)
	{
		case DP_DSP | DC_ETHERNET:
		case DP_UNKNOWN | DC_ETHERNET:
			data_type = -5;
		break;
		CASE_DSP (data_type, GetDataType, &dev, fd, dataID)
		CASE_USB (data_type, GetDataType, &dev, fd, dataID)
		CASE_TRANSPUTER (data_type, GetDataType, &dev, fd, dataID)
		default:
			adwin_errno = -EINVAL;
			data_type = -1;
	}

	switch(data_type) {
		case 2:
			ptr = malloc(count * 2);
			break;
		case 3:
		case 4:
			ptr = malloc(count * 4);
			break;
		case 5:
		case -5:
			ptr = dest;
			break;

		default:
			CLEARMUTEX(&dev, fd);
			return 255;
	}

	if (!ptr ) {
		adwin_errno = -errno;
		return 255;
	}

	if ((data_type == 3 || data_type == 4)
	    && dev.dev_type == DC_ETHERNET)
		data_type = 2;
	SELECT_PROTOCOL_DATA (Get, res, &dev, fd, dataID, first, count, data_type == -5 ? 5 : data_type, ptr);
	CLEARMUTEX(&dev, fd);

	switch (data_type) {
		case 2:		/* short */
			if (! res)
				CAST(int16_t, ptr, float, dest, count);
			free(ptr);
			break;
		case 3:		/* integer */
		case 4:		/* long */
			if (! res)
				CAST(int32_t, ptr, float, dest, count);
			free(ptr);
			break;
		case 5:
			break;
	}

	return res ? 255 : 0;
}

/*****************************************************************************/
void SetData_Float(long DataNo, float *Data, long StartIndex, long Count){
	adwin_set_data_float (DataNo, StartIndex, Count, Data, CURRENTDEVICE.DeviceNo);
}

int adwin_set_data_float (int dataID, long first, long count, float *src, int id) {
	adwin_dev_t dev;
	int fd, data_type, res;
	void *ptr=NULL;

	data_type = adwin_get_data_type(dataID, id);
	switch (data_type) {
		case 2:
			ptr = malloc(count * 2);
			CAST(float, src, int16_t, ptr, count);
			break;
		case 3:
		case 4:
			ptr = malloc(count * 4);
			CAST(float, src, int32_t, ptr, count);
			break;
		case 5:
			ptr = src;
			break;
		default:
			return 255;
	}

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0) {
		if (ptr != src) free(ptr);
		return 255;
	}

	if ((data_type == 3 || data_type == 4)
	    && dev.dev_type == DC_ETHERNET)
		data_type = 2;
	SELECT_PROTOCOL_DATA (Set, res, &dev, fd, dataID, first, count, data_type, ptr);
	CLEARMUTEX(&dev, fd);

	if (ptr != src) free(ptr);
	return res ? 255 : 0;
}


/*****************************************************************************/
void GetData_Double(long DataNo, double *Data, long StartIndex, long Count){
	adwin_get_data_double (DataNo, StartIndex, Count, Data, CURRENTDEVICE.DeviceNo);
}

int adwin_get_data_double(int dataID, long first, long count, double *dest, int id) {
	int res;
	float *ptr;

	ptr = malloc(count * 4);
	res = adwin_get_data_float(dataID, first, count, ptr, id);
	// Use FCAST only when casting from floating point numbers to integer types
	// (FCAST does the same as CAST, except that it additionally converts NaN-values to the
	//  largest integer compatible with the target type.
	//  When using floating point types as targets, this conversion is not necessary)
	CAST(float, ptr, double, dest, count);
	free(ptr);
	return res;
}

/*****************************************************************************/
void SetData_Double(long DataNo, double *Data, long StartIndex, long Count){
	adwin_set_data_double(DataNo, StartIndex, Count, Data, CURRENTDEVICE.DeviceNo);
}

int adwin_set_data_double(int dataID, long first, long count, double *src, int id) {
	int res;
	float *ptr;

	ptr = malloc(count * 4);
	CAST(double, src, float, ptr, count);
	res = adwin_set_data_float(dataID, first, count, ptr, id);
	free(ptr);
	return res;
}


/*****************************************************************************/
long GetData_String (long DataNo, char *Data, long MaxCount){
	return adwin_get_data_string(DataNo, MaxCount, Data, CURRENTDEVICE.DeviceNo);
}

int adwin_get_data_string(int data, long count, char *dest, int id) {
	int32_t i, len, *ptr;

	ptr = (int32_t *) malloc((count + 2) * sizeof(int32_t));

	if (adwin_get_data_int32(data, 1, count + 1, ptr, id) != 0) {
		free(ptr);
		return -1;
	}

	len = ptr[0];
	if (count < len)
		len = count;

	for (i = 0; i < len; i ++)
		dest[i] = (char) ptr[i + 1];

	dest[len] = '\0';

	free(ptr);
	return len;
}

/*****************************************************************************/
void SetData_String (long DataNo, char *Data){
	 adwin_set_data_string(DataNo, Data, CURRENTDEVICE.DeviceNo);
}

int adwin_set_data_string(int data, const char *src, int id) {
	int32_t *ptr;
	long tmp, i, len;

	len = strlen(src);
	ptr = (int32_t *) malloc((len  + 2) * sizeof(int32_t));

	ptr[0] = len;

	for (i = 0; i < len + 1; i ++)
		ptr[i + 1] = (unsigned char) src[i];

	ptr[len + 1] = 0;

	tmp = adwin_set_data_int32(data, 1, 1 + len + 1, ptr, id);
	free(ptr);

	return (tmp != 0) ? -1 : len;
}

/*****************************************************************************/
long String_Length (long DataNo){
	return adwin_strlen(DataNo, CURRENTDEVICE.DeviceNo);
}

long adwin_strlen(int data, int id) {
	int len;
	return adwin_get_data_int32(data, 1, 1, &len, id) != 0 ? -1 : len;
}

/*****************************************************************************/
long Fifo_Full (long FifoNo){
	return adwin_fifo_full(FifoNo, CURRENTDEVICE.DeviceNo);
}

int adwin_fifo_full (int fifo, int id) {
	int cnt, err;
	cnt = __adwin_fifo_full(fifo, &err, id);
	return (err == -1) ? 255 : cnt;
}


int __adwin_fifo_full (int fifo, int *err, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;
	int dest;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL_WITH_RETURN1(tmp, Fifo_Full, &dev, fd, fifo, &dest);
	CLEARMUTEX(&dev, fd);

	if (err)
		*err = tmp;

	return dest;
}

/*****************************************************************************/
long Fifo_Empty (long FifoNo){
	return adwin_fifo_empty(FifoNo, CURRENTDEVICE.DeviceNo);
}

int adwin_fifo_empty (int fifo, int id) {
	int err, cnt;
	cnt = __adwin_fifo_empty(fifo, &err, id);

	return ((err == -1) ? 255 : cnt);
}

int __adwin_fifo_empty (int fifo, int *err, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;
	int dest;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL_WITH_RETURN1(tmp, Fifo_Empty, &dev, fd, fifo, &dest);
	CLEARMUTEX(&dev, fd);

	if (err)
		*err = tmp;

	return dest;
}

/*****************************************************************************/
void Fifo_Clear (long FifoNo){
	adwin_fifo_clear(FifoNo, CURRENTDEVICE.DeviceNo);
}

int adwin_fifo_clear(int fifoID, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Fifo_Clear, &dev, fd, fifoID);

	CLEARMUTEX(&dev, fd);
	return (tmp != -1) ? tmp : 255;
}

/*****************************************************************************/
int adwin_get_fifo_int16(int fifo, long count, int16_t *dest, int id) {
	int res;
	int32_t *ptr;

	ptr = malloc(count * 4);
	res = adwin_get_fifo_int32(fifo, count, ptr, id);
	CAST(int32_t, ptr, int16_t, dest, count);
	free(ptr);
	return res;
}

/*****************************************************************************/
void GetFifo_Long (long FifoNo, long *Data, long Count){
	adwin_get_fifo_int32(FifoNo, Count, (int32_t*)Data, CURRENTDEVICE.DeviceNo);
}

int adwin_get_fifo_int32(int fifo, long count, int32_t *dest, int id) {
	adwin_dev_t dev;
	int fd, res, data_type;
	void *ptr;

	data_type = adwin_get_data_type(fifo, id);
	switch(data_type) {
		case 2:
			ptr = malloc(count * 2);
			break;
		case 3:
		case 4:
			ptr = dest;
			break;
		case 5:
			ptr = malloc(count * 4);
			break;
		default:
			return 255;
	}

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0) {
		if (ptr != dest) free (ptr);
		return 255;
	}

	// awkward: when receiving the data type, 3 means int and 4 means long.
	// When sending, 2 means int, but only
	// on ethernet devices (on link devices, 2 always means short)
	if ((data_type == 3 || data_type == 4)
	    && dev.dev_type == DC_ETHERNET)
		data_type = 2;
	SELECT_PROTOCOL_FIFO (Get, res, &dev, fd, fifo, count, data_type, ptr);

	if (data_type == 2 && dev.dev_type == DC_ETHERNET)
		data_type = 3;

	CLEARMUTEX(&dev, fd);

	switch (data_type) {
		case 2:
			if (! res)
				CAST(int16_t, ptr, int32_t, dest, count);
			free(ptr);
			break;

		case 3:
		case 4:
			break;

		case 5:
			if (! res)
				FCAST(float, ptr, int32_t, dest, count);
			free(ptr);
			break;
	}

	return res ? 255 : 0;
}

/*****************************************************************************/
int adwin_get_fifo_int64(int fifo, long count, int64_t *dest, int id) {
	int res;
	int32_t *ptr;

	ptr = malloc(count * 4);
	res = adwin_get_fifo_int32(fifo, count, ptr, id);
	CAST(int32_t, ptr, int64_t, dest, count);
	free(ptr);
	return res;
}

/*****************************************************************************/
int adwin_set_fifo_int16(int fifo, long count, int16_t *src, int id) {
	int res;
	int32_t *ptr;

	ptr = malloc(count * 4);
	CAST(int16_t, src, int32_t, ptr, count);
	res = adwin_set_fifo_int32(fifo, count, ptr, id);
	free(ptr);
	return res;
}

/*****************************************************************************/
void SetFifo_Long (long FifoNo, long *Data, long Count) {
	adwin_set_fifo_int32(FifoNo, Count, (int32_t*)Data, CURRENTDEVICE.DeviceNo);
}

int adwin_set_fifo_int32(int fifo, long count, int32_t *src, int id) {
	adwin_dev_t dev;
	int fd, data_type, res;
	void *ptr;

	data_type = adwin_get_data_type(fifo, id);
	switch(data_type) {
		case 2:
			ptr = malloc(count * 2);
			CAST(int32_t, src, int16_t, ptr, count);
			break;
		case 3:
		case 4:
			ptr = src;
			break;
		case 5:
			ptr = malloc(count * 4);
			CAST(int32_t, src, float, ptr, count);
			break;
		default:
			return 255;
	}

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0) {
		if (ptr != src) free(ptr);
		return 255;
	}

	if ((data_type == 3 || data_type == 4)
	    && dev.dev_type == DC_ETHERNET)
		data_type = 2;
	SELECT_PROTOCOL_FIFO (Set, res, &dev, fd, fifo, count, data_type, ptr);
	CLEARMUTEX(&dev, fd);

	if (ptr != src) free(ptr);
	return res ? 255 : 0;
}

/*****************************************************************************/

int adwin_set_fifo_int64(int fifo, long count, int64_t *src, int id) {
	int res;
	int32_t *ptr;

	ptr = malloc(count * 4);
	CAST(int64_t, src, int32_t, ptr, count);
	res = adwin_set_fifo_int32(fifo, count, ptr, id);
	free(ptr);
	return res;
}

/*****************************************************************************/
void GetFifo_Float (long FifoNo, float *Data, long Count){
	adwin_get_fifo_float(FifoNo, Count, Data, CURRENTDEVICE.DeviceNo);
}

int adwin_get_fifo_float(int fifo, long count, float *dest, int id) {
	adwin_dev_t dev;
	int fd, res, data_type;
	void *ptr;

	data_type = adwin_get_data_type(fifo, id);
	switch(data_type) {
		case 2:
			ptr = malloc(count * 2);
			break;
		case 3:
		case 4:
			ptr = malloc(count * 4);
			break;
		case 5:
			ptr = dest;
			break;

		default:
			return 255;
	}

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0) {
		if (ptr != dest) free(ptr);
		return 255;
	}

	if ((data_type == 3 || data_type == 4)
	    && dev.dev_type == DC_ETHERNET)
		data_type = 2;
	SELECT_PROTOCOL_FIFO (Get, res, &dev, fd, fifo, count, data_type, ptr);
	if (data_type == 2 && dev.dev_type == DC_ETHERNET) data_type = 3;
	CLEARMUTEX(&dev, fd);

	switch (data_type) {
		case 2:		/* short */
			if (! res)
				CAST(int16_t, ptr, float, dest, count);
			free(ptr);
			break;
		case 3:		/* integer */
		case 4:		/* long */
			if (! res)
				CAST(int32_t, ptr, float, dest, count);
			free(ptr);
			break;
		case 5:
			break;
	}

	return res ? 255 : 0;
}

/*****************************************************************************/
void SetFifo_Float (long FifoNo, float *Data, long Count) {
	adwin_set_fifo_float(FifoNo, Count, Data, CURRENTDEVICE.DeviceNo);
}

int adwin_set_fifo_float(int fifo, long count, float *src, int id) {
	adwin_dev_t dev;
	int fd, data_type, res;
	void *ptr;

	data_type = adwin_get_data_type(fifo, id);
	switch(data_type) {
		case 2:
			ptr = malloc(count * 2);
			CAST(float, src, int16_t, ptr, count);
			break;
		case 3:
		case 4:
			ptr = malloc(count * 4);
			CAST(float, src, int32_t, ptr, count);
			break;
		case 5:
			ptr = src;
			break;

		default:
			return 255;
	}

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0) {
		if (ptr != src) free(ptr);
		return 255;
	}


	if ((data_type == 3 || data_type == 4)
	    && dev.dev_type == DC_ETHERNET)
		data_type = 2;
	SELECT_PROTOCOL_FIFO (Set, res, &dev, fd, fifo, count, data_type, ptr);
	CLEARMUTEX(&dev, fd);

	if (ptr != src) free(ptr);
	return res ? 255 : 0;
}

/*****************************************************************************/
void GetFifo_Double (long FifoNo, double *Data, long Count){
	adwin_get_fifo_double(FifoNo, Count, Data, CURRENTDEVICE.DeviceNo);
}

int adwin_get_fifo_double(int fifo, long count, double *dest, int id) {
	int res;
	float *ptr;

	ptr = malloc(count * 4);
	res = adwin_get_fifo_float(fifo, count, ptr, id);
	CAST(float, ptr, double, dest, count);
	free(ptr);
	return res;
}

/*****************************************************************************/
void SetFifo_Double (long FifoNo, double *Data, long Count){
	adwin_set_fifo_double(FifoNo, Count, Data, CURRENTDEVICE.DeviceNo);
}

int adwin_set_fifo_double(int fifo, long count, double *src, int id) {
	int res;
	float *ptr;

	ptr = malloc(count * 4);
	CAST(double, src, float, ptr, count);
	res = adwin_set_fifo_float(fifo, count, ptr, id);
	free(ptr);
	return res;
}

/*****************************************************************************/
long Get_Globaldelay(long ProcessNo){
	return adwin_get_globaldelay(ProcessNo, CURRENTDEVICE.DeviceNo);
}

int adwin_get_globaldelay(int process, int id) {
	return adwin_get_par(process - 90, id);
}

int __adwin_get_globaldelay(int process, int *err, int id) {
	return __adwin_get_par(process - 90, err, id);
}

/*****************************************************************************/
long Set_Globaldelay(long ProcessNo, long Value){
	return adwin_set_globaldelay(ProcessNo, Value, CURRENTDEVICE.DeviceNo);
}

int adwin_set_globaldelay(int process, int delay, int id) {
	return adwin_set_par(process - 90, delay, id);
}

/*****************************************************************************/
//long Processor_Type(){
//	return adwin_processor_type(DeviceNo);
//}

long adwin_processor_type(int id) {
#warning adwin_processor_type() disabled!
/*	adwin_dev_t dev;
	char proc[255];
	int fd;

	sprintf(proc, "/proc/adwin/%d/0x%X", getpid(), id);
	if ((fd = open(proc, O_RDONLY | O_EXCL)) < 0) {
		errno = ENOENT;
		return -1;
	}

	ioctl(fd, ADIOGKDEVINFO, &dev);
	close(fd);

	switch (dev.dev_processor) {
		case DP_21160:
		case DP_DSP:	return 9;
		case DP_T2:		return 2;
		case DP_T4:		return 4;
		case DP_T8:		return 8;
		default:			return 0;
	}*/
	return 0;
}

/*****************************************************************************/
/*
 * mode 0:	vorhandene Datei berschreiben
 * mode 1:	wenn Datei existiert, Daten anhngen
 *
 * binr schreiben, ohne Header
 *
 * Daniel:
 * wenigstens DataNo und von-bis knnte man dazuschreiben, evtl. mit Datum ?!
 */

void Data2File (char *Filename, long DataNo, long Starindex, long Count, int Mode){
	adwin_data2file(Filename, DataNo, Starindex, Count, Mode, CURRENTDEVICE.DeviceNo);
}

int adwin_data2file(char *file, int data, long first, long cnt, int mode, int id) {
	int fd;
	long nbytes;
	char *ptr = NULL;

	switch(adwin_get_data_type(data, id)) {
		case 2:
			errno = ENOTSUP;
			return -1;
		case 3:
		case 4:
			nbytes = cnt *4;
			ptr = (char *) malloc(nbytes);
			if (adwin_get_data_int32(data, first, cnt, (int32_t *)ptr, id)) {
				free(ptr);
				return -1;
			}
			break;
		case 5:
			nbytes = cnt *4;
			ptr = (char *) malloc(nbytes);
			if (adwin_get_data_float(data, first, cnt, (float *)ptr, id)) {
				free(ptr);
				return -1;
			}
			break;
		default:
			return -1;
	}

	errno = 0;
	//if ((fd = open(file, O_WRONLY | O_CREAT | (mode) ? O_APPEND : O_TRUNC ), S_IRWXU | S_IRWXG | S_IRWXO) < 0) {
	//#warning Daniel: O_CREAT und O_EXCL NICHT zusammen angeben, data2file() sollte jetzt gehen !?
	fd = open(file, O_WRONLY | O_CREAT | (mode ? O_APPEND : O_TRUNC), 0644);
	if (fd < 0) {
		free(ptr);
		perror("open");
		return -1;
	}

	errno = 0;
	if (write(fd, ptr, nbytes) < nbytes) {
		perror ("write");
		free(ptr);
		return -1;
	}

	free(ptr);
	close(fd);

	return 0;
}

/*****************************************************************************/
static struct {
	int __errno;
	char *__errstr;
} adwin_errlist[] = {
	{ k_timeout_to_link, 		"Timeout error on writing to the ADwin-system." },
	{ k_timeout_from_link,		"Timeout error on reading from the ADwin-system."},
	{ k_this_link_not_allowed,	"The device No. is not allowed."},
	{ k_this_link_is_unknown,	"The device No. is not known."},
	{ k_unknown_device,		"Device is unknown."},
	{ k_function_not_allowed,	"Function for this device not allowed." },
	{ k_wrong_version,		"Incompatible versions of ADwin operating system ,\ndriver (ADwin32.DLL) and/or ADbasic binary-file." },
	{ k_Data_is_to_small,		"The Data is too small." },
	{ k_Fifo_is_to_small,		"The Fifo is too small or not enough values." },
	{ k_Fifo_to_less_data,		"The Fifo has not enough values." },
	{ k_File_not_found,		"File not found." },
	{ k_File_not_binaer,		"The file is not an ADBasic binary-file." },
	{ k_File_not_valid,		"The file is not valid." },
	{ k_File_not_btl,		"The file is not a BTL." },
	{ k_error_net_error,		"Network error (RPC)." },
	{ k_error_net_binding_compose,	"Network error (Bind)." },
	{ k_error_net_stringbinding,	"Network error (String bind)." },
	{ k_error_net_password,		"Wrong password (RPC)." },
	{ k_error_tcp_error,		"Network error (RPC)." },
	{ k_error_tcp_timout_read,	"Network timeout." },
	{ k_error_tcp_password,		"Wrong password." },
	{ 0, NULL},
};

const char *adwin_strerror(int e) {
	int i;

	if (e == 0)
		return "No Error.";

	if (e < 0)
		return strerror(-e);

	for (i = 0; adwin_errlist[i].__errstr; i++)
		if (adwin_errlist[i].__errno == e)
			return adwin_errlist[i].__errstr;

	return "No description available";
}

void adwin_perror(const char *msg) {
	fprintf(stderr, "%s: %s\n", msg, adwin_strerror(adwin_errno));
}



short Get_Data_packed (void *tarr, short typ, short nadc, long lstart, long il ,short nadr){
	/*   typ = 1 fr short
	     typ = 2 fr long
	     typ = 5 fr float
	     typ = 6 fr double  */

     switch (typ)
     {
        case 1:  //short
	{
	    adwin_errno=k_function_not_allowed;
	    return 255;
	}
     	case 2:  //long
	{
	    adwin_errno=k_function_not_allowed;
	    return 255;
	}
	case 5:  //float
	{
	    adwin_errno=k_function_not_allowed;
	    return 255;
	}
	case 6:  //double
	{
	    adwin_errno=k_function_not_allowed;
	    return 255;
	}

	default :
	{
            adwin_errno=k_function_not_allowed;
	    return 255;
	}

     }
}

long Get_Last_Error ()
{
	long temp = adwin_errno;
	adwin_errno = 0;
	return temp;
}

char *Get_Last_Error_Text (long Errno)
{
	return (char*)(adwin_strerror (Errno));
}


// 30.07.04 KS:

#define DEBUG_VERSION 1

int32_t get_internal_version ()
{
	char *version = VERSION;
	char v[40];
	int vi = 0;
	int n;

	for (n = 0; version[n]; n++) if (version[n] >= '0' && version[n] <= '9') v[vi++] = version[n];
	v[vi] = 0;
	return strtol (v, NULL, 10) | (DEBUG_VERSION << 24);
}
